@@ -74,6 +74,7 @@ FLOT = 'flot'
 
 ABSOLUTE_TYPE = 'absolute'
 DASHBOARD_TYPE = 'dashboard'
+ROW_TYPE = 'row'
 GRAPH_TYPE = 'graph'
 STAT_TYPE = 'stat'
 SINGLESTAT_TYPE = 'singlestat'
@@ -83,6 +84,7 @@ ALERTLIST_TYPE = 'alertlist'
 BARGAUGE_TYPE = 'bargauge'
 GAUGE_TYPE = 'gauge'
 DASHBOARDLIST_TYPE = 'dashlist'
+LOGS_TYPE = 'logs'
 HEATMAP_TYPE = 'heatmap'
 STATUSMAP_TYPE = 'flant-statusmap-panel'
 SVG_TYPE = 'marcuscalidus-svg-panel'
@@ -538,46 +540,28 @@ def _balance_panels(panels):
 
 
 @attr.s
-class Row(object):
-    # TODO: jml would like to separate the balancing behaviour from this
-    # layer.
-    panels = attr.ib(default=attr.Factory(list), converter=_balance_panels)
-    collapse = attr.ib(
-        default=False, validator=instance_of(bool),
-    )
-    editable = attr.ib(
-        default=True, validator=instance_of(bool),
-    )
-    height = attr.ib(
-        default=attr.Factory(lambda: DEFAULT_ROW_HEIGHT),
-        validator=instance_of(Pixels),
-    )
-    showTitle = attr.ib(default=None)
-    title = attr.ib(default=None)
-    repeat = attr.ib(default=None)
+class GridPos(object):
+    """GridPos describes the panel size and position in grid coordinates.
 
-    def _iter_panels(self):
-        return iter(self.panels)
+        :param h: height of the panel, grid height units each represents
+            30 pixels
+        :param w: width of the panel 1-24 (the width of the dashboard
+            is divided into 24 columns)
+        :param x: x cordinate of the panel, in same unit as w
+        :param y: y cordinate of the panel, in same unit as h
+    """
 
-    def _map_panels(self, f):
-        return attr.evolve(self, panels=list(map(f, self.panels)))
+    h = attr.ib()
+    w = attr.ib()
+    x = attr.ib()
+    y = attr.ib()
 
     def to_json_data(self):
-        showTitle = False
-        title = "New row"
-        if self.title is not None:
-            showTitle = True
-            title = self.title
-        if self.showTitle is not None:
-            showTitle = self.showTitle
         return {
-            'collapse': self.collapse,
-            'editable': self.editable,
-            'height': self.height,
-            'panels': self.panels,
-            'showTitle': showTitle,
-            'title': title,
-            'repeat': self.repeat,
+            'h': self.h,
+            'w': self.w,
+            'x': self.x,
+            'y': self.y
         }
 
 
@@ -817,13 +801,24 @@ DEFAULT_TIME = Time('now-1h', 'now')
 
 @attr.s
 class TimePicker(object):
+    """
+    Time Picker
+    :param refreshIntervals: dashboard auto-refresh interval options
+    :param timeOptions: dashboard time range options
+    :param hidden: hide the time picker from dashboard
+    """
     refreshIntervals = attr.ib()
     timeOptions = attr.ib()
+    hidden = attr.ib(
+        default=False,
+        validator=instance_of(bool),
+    )
 
     def to_json_data(self):
         return {
             'refresh_intervals': self.refreshIntervals,
             'time_options': self.timeOptions,
+            'hidden': self.hidden
         }
 
 
@@ -928,7 +923,7 @@ class AlertCondition(object):
     timeRange = attr.ib(validator=instance_of(TimeRange))
     operator = attr.ib()
     reducerType = attr.ib()
-    type = attr.ib(default=CTYPE_QUERY)
+    type = attr.ib(default=CTYPE_QUERY, kw_only=True)
 
     def to_json_data(self):
         queryParams = [
@@ -978,11 +973,21 @@ class Alert(object):
         }
 
 
+@attr.s
+class Notification(object):
+
+    uid = attr.ib()
+
+    def to_json_data(self):
+        return {
+            'uid': self.uid,
+        }
+
+
 @attr.s
 class Dashboard(object):
 
     title = attr.ib()
-    rows = attr.ib()
     annotations = attr.ib(
         default=attr.Factory(Annotations),
         validator=instance_of(Annotations),
@@ -1000,7 +1005,9 @@ class Dashboard(object):
     id = attr.ib(default=None)
     inputs = attr.ib(default=attr.Factory(list))
     links = attr.ib(default=attr.Factory(list))
+    panels = attr.ib(default=attr.Factory(list), validator=instance_of(list))
     refresh = attr.ib(default=DEFAULT_REFRESH)
+    rows = attr.ib(default=attr.Factory(list), validator=instance_of(list))
     schemaVersion = attr.ib(default=SCHEMA_VERSION)
     sharedCrosshair = attr.ib(
         default=False,
@@ -1029,8 +1036,20 @@ class Dashboard(object):
             for panel in row._iter_panels():
                 yield panel
 
+        for panel in self.panels:
+            if hasattr(panel, 'panels'):
+                yield panel
+                for row_panel in panel._iter_panels():
+                    yield panel
+            else:
+                yield panel
+
     def _map_panels(self, f):
-        return attr.evolve(self, rows=[r._map_panels(f) for r in self.rows])
+        return attr.evolve(
+            self,
+            rows=[r._map_panels(f) for r in self.rows],
+            panels=[p._map_panels(f) for p in self.panels]
+        )
 
     def auto_panel_ids(self):
         """Give unique IDs all the panels without IDs.
@@ -1048,6 +1067,11 @@ class Dashboard(object):
         return self._map_panels(set_id)
 
     def to_json_data(self):
+        if self.panels and self.rows:
+            print(
+                "Warning: You are using both panels and rows in this dashboard, please use one or the other. "
+                "Panels should be used in preference over rows, see example dashboard for help."
+            )
         return {
             '__inputs': self.inputs,
             'annotations': self.annotations,
@@ -1057,6 +1081,7 @@ class Dashboard(object):
             'hideControls': self.hideControls,
             'id': self.id,
             'links': self.links,
+            'panels': self.panels if not self.rows else [],
             'refresh': self.refresh,
             'rows': self.rows,
             'schemaVersion': self.schemaVersion,
@@ -1105,55 +1130,144 @@ class Panel(object):
     editable = attr.ib(default=True, validator=instance_of(bool))
     error = attr.ib(default=False, validator=instance_of(bool))
     height = attr.ib(default=None)
+    gridPos = attr.ib(default=None)
     hideTimeOverride = attr.ib(default=False, validator=instance_of(bool))
     id = attr.ib(default=None)
     interval = attr.ib(default=None)
     links = attr.ib(default=attr.Factory(list))
     maxDataPoints = attr.ib(default=100)
     minSpan = attr.ib(default=None)
-    repeat = attr.ib(default=None)
+    repeat = attr.ib(default=attr.Factory(Repeat), validator=instance_of(Repeat))
     span = attr.ib(default=None)
     timeFrom = attr.ib(default=None)
     timeShift = attr.ib(default=None)
     transparent = attr.ib(default=False, validator=instance_of(bool))
 
+    def _map_panels(self, f):
+        return f(self)
+
     def panel_json(self, overrides):
         res = {
-            "cacheTimeout": self.cacheTimeout,
-            "datasource": self.dataSource,
-            "description": self.description,
-            "editable": self.editable,
-            "error": self.error,
-            "height": self.height,
-            "hideTimeOverride": self.hideTimeOverride,
-            "id": self.id,
-            "interval": self.interval,
-            "links": self.links,
-            "maxDataPoints": self.maxDataPoints,
-            "minSpan": self.minSpan,
-            "repeat": self.repeat,
-            "span": self.span,
-            "targets": self.targets,
-            "timeFrom": self.timeFrom,
-            "timeShift": self.timeShift,
-            "title": self.title,
-            "transparent": self.transparent,
+            'cacheTimeout': self.cacheTimeout,
+            'datasource': self.dataSource,
+            'description': self.description,
+            'editable': self.editable,
+            'error': self.error,
+            'height': self.height,
+            'gridPos': self.gridPos,
+            'hideTimeOverride': self.hideTimeOverride,
+            'id': self.id,
+            'interval': self.interval,
+            'links': self.links,
+            'maxDataPoints': self.maxDataPoints,
+            'minSpan': self.minSpan,
+            'repeat': self.repeat.variable,
+            'repeatDirection': self.repeat.direction,
+            'maxPerRow': self.repeat.maxPerRow,
+            'span': self.span,
+            'targets': self.targets,
+            'timeFrom': self.timeFrom,
+            'timeShift': self.timeShift,
+            'title': self.title,
+            'transparent': self.transparent,
         }
         res.update(overrides)
         return res
 
 
+@attr.s
+class RowPanel(Panel):
+    """
+    Generates Row panel json structure.
+
+    :param title: title of the panel
+    :param panels: list of panels in the row
+    """
+
+    panels = attr.ib(default=attr.Factory(list), validator=instance_of(list))
+
+    def _iter_panels(self):
+        return iter(self.panels)
+
+    def _map_panels(self, f):
+        self = f(self)
+        return attr.evolve(self, panels=list(map(f, self.panels)))
+
+    def to_json_data(self):
+        return self.panel_json(
+            {
+                'collapsed': False,
+                'panels': self.panels,
+                'type': ROW_TYPE
+            }
+        )
+
+
+@attr.s
+class Row(object):
+    """
+    Legacy support for old row, when not used with gridpos
+    """
+    # TODO: jml would like to separate the balancing behaviour from this
+    # layer.
+    try:
+        panels = attr.ib(default=attr.Factory(list), converter=_balance_panels)
+    except TypeError:
+        panels = attr.ib(default=attr.Factory(list), convert=_balance_panels)
+    collapse = attr.ib(
+        default=False, validator=instance_of(bool),
+    )
+    editable = attr.ib(
+        default=True, validator=instance_of(bool),
+    )
+    height = attr.ib(
+        default=attr.Factory(lambda: DEFAULT_ROW_HEIGHT),
+        validator=instance_of(Pixels),
+    )
+    showTitle = attr.ib(default=None)
+    title = attr.ib(default=None)
+    repeat = attr.ib(default=None)
+
+    def _iter_panels(self):
+        return iter(self.panels)
+
+    def _map_panels(self, f):
+        return attr.evolve(self, panels=list(map(f, self.panels)))
+
+    def to_json_data(self):
+        showTitle = False
+        title = "New row"
+        if self.title is not None:
+            showTitle = True
+            title = self.title
+        if self.showTitle is not None:
+            showTitle = self.showTitle
+        return {
+            'collapse': self.collapse,
+            'editable': self.editable,
+            'height': self.height,
+            'panels': self.panels,
+            'showTitle': showTitle,
+            'title': title,
+            'repeat': self.repeat,
+        }
+
+
 @attr.s
 class Graph(Panel):
     """
     Generates Graph panel json structure.
 
-    :param dataLinks: list of data links hooked to datapoints on the graph
+    :param alert: List of AlertConditions
+    :param alertRuleTags: Key Value pairs to be sent with Alert notifications
+    :param dataLinks: List of data links hooked to datapoints on the graph
     :param dataSource: DataSource's name
     :param minSpan: Minimum width for each panel
-    :param repeat: Template's name to repeat Graph on
+    :param thresholds: List of GraphThresholds - Only valid when alert not defined
     """
 
+    alert = attr.ib(default=None)
+    alertRuleTags = attr.ib(default=attr.Factory(dict))
     alertThreshold = attr.ib(default=True, validator=instance_of(bool))
     aliasColors = attr.ib(default=attr.Factory(dict))
     bars = attr.ib(default=False, validator=instance_of(bool))
@@ -1180,14 +1294,21 @@ class Graph(Panel):
         default=attr.Factory(Tooltip),
         validator=instance_of(Tooltip),
     )
+    thresholds = attr.ib(default=attr.Factory(list))
     xAxis = attr.ib(default=attr.Factory(XAxis), validator=instance_of(XAxis))
     # XXX: This isn't a *good* default, rather it's the default Grafana uses.
-    yAxes = attr.ib(
-        default=attr.Factory(YAxes),
-        converter=to_y_axes,
-        validator=instance_of(YAxes),
-    )
-    alert = attr.ib(default=None)
+    try:
+        yAxes = attr.ib(
+            default=attr.Factory(YAxes),
+            converter=to_y_axes,
+            validator=instance_of(YAxes),
+        )
+    except TypeError:
+        yAxes = attr.ib(
+            default=attr.Factory(YAxes),
+            convert=to_y_axes,
+            validator=instance_of(YAxes),
+        )
 
     def to_json_data(self):
         graphObject = {
@@ -1214,12 +1335,17 @@ class Graph(Panel):
             'stack': self.stack,
             'steppedLine': self.steppedLine,
             'tooltip': self.tooltip,
+            'thresholds': self.thresholds,
             'type': GRAPH_TYPE,
             'xaxis': self.xAxis,
             'yaxes': self.yAxes,
         }
         if self.alert:
             graphObject['alert'] = self.alert
+            graphObject['alertRuleTags'] = self.alertRuleTags
+            graphObject['thresholds'] = []
+        if self.thresholds and self.alert:
+            print("Warning: Graph threshold ignored as Alerts defined")
         return self.panel_json(graphObject)
 
     def _iter_targets(self):
@@ -1227,7 +1353,7 @@ class Graph(Panel):
             yield target
 
     def _map_targets(self, f):
-        return attr.assoc(self, targets=[f(t) for t in self.targets])
+        return attr.evolve(self, targets=[f(t) for t in self.targets])
 
     def auto_ref_ids(self):
         """Give unique IDs all the panels without IDs.
@@ -1249,7 +1375,7 @@ class Graph(Panel):
         auto_ref_ids = (i for i in candidate_ref_ids if i not in ref_ids)
 
         def set_refid(t):
-            return t if t.refId else attr.assoc(t, refId=next(auto_ref_ids))
+            return t if t.refId else attr.evolve(t, refId=next(auto_ref_ids))
         return self._map_targets(set_refid)
 
 
@@ -1404,7 +1530,6 @@ class Stat(Panel):
     :param span: defines the number of spans that will be used for panel
     :param thresholds: single stat thresholds
     :param transparent: defines if the panel should be transparent
-    :param repeat: defines how the panel should be repeated
     """
 
     textMode = attr.ib(default='auto')
@@ -1418,7 +1543,6 @@ class Stat(Panel):
     thresholds = attr.ib(default="")
     reduceCalc = attr.ib(default='mean', type=str)
     decimals = attr.ib(default=None)
-    repeat = attr.ib(default=attr.Factory(Repeat), validator=instance_of(Repeat))
 
     def to_json_data(self):
         return self.panel_json(
@@ -1450,9 +1574,6 @@ class Stat(Panel):
                     }
                 },
                 'type': STAT_TYPE,
-                'repeat': self.repeat.variable,
-                'repeatDirection': self.repeat.direction,
-                'maxPerRow': self.repeat.maxPerRow
             }
         )
 
@@ -2285,7 +2406,8 @@ class StatusmapColor(object):
 
 @attr.s
 class Statusmap(Panel):
-    """Generates json structure for the flant-statusmap-panel visualisation plugin (https://grafana.com/grafana/plugins/flant-statusmap-panel/).
+    """Generates json structure for the flant-statusmap-panel visualisation plugin
+    (https://grafana.com/grafana/plugins/flant-statusmap-panel/).
 
     :param alert
     :param cards: A statusmap card object: keys 'cardRound', 'cardMinWidth', 'cardHSpacing', 'cardVSpacing'
@@ -2377,7 +2499,8 @@ class Svg(Panel):
     :param id: panel id
     :param interval: defines time interval between metric queries
     :param links: additional web links
-    :param reduceCalc: algorithm for reduction to a single value: keys 'mean' 'lastNotNull' 'last' 'first' 'firstNotNull' 'min' 'max' 'sum' 'total'
+    :param reduceCalc: algorithm for reduction to a single value:
+        keys 'mean' 'lastNotNull' 'last' 'first' 'firstNotNull' 'min' 'max' 'sum' 'total'
     :param span: defines the number of spans that will be used for panel
     :param svgFilePath: path to SVG image file to be displayed
     """
@@ -2479,7 +2602,8 @@ class DashboardList(Panel):
     :param transparent: defines if the panel is transparent
 
     :param showHeadings: The chosen list selection (Starred, Recently viewed, Search) is shown as a heading
-    :param showSearch: Display dashboards by search query or tags. Requires you to enter at least one value in Query or Tags
+    :param showSearch: Display dashboards by search query or tags.
+        Requires you to enter at least one value in Query or Tags
     :param showRecent: Display recently viewed dashboards in alphabetical order
     :param showStarred: Display starred dashboards in alphabetical order
     :param maxItems: Sets the maximum number of items to list per section
@@ -2497,7 +2621,6 @@ class DashboardList(Panel):
     def to_json_data(self):
         return self.panel_json(
             {
-                'height': self.height,
                 'fieldConfig': {
                     'defaults': {
                         'custom': {},
@@ -2517,12 +2640,66 @@ class DashboardList(Panel):
 
 
 @attr.s
-class Threshold(object):
-    """Threshold for a gauge
+class Logs(Panel):
+    """Generates Logs panel json structure
+    Grafana doc on Logs panel: https://grafana.com/docs/grafana/latest/panels/visualizations/logs-panel/
+    :param title: panel title
+    :param description: optional panel description
+    :param editable: defines if panel is editable via web interfaces
+    :param height: defines panel height
+    :param id: panel id
+    :param links: additional web links
+    :param span: defines the number of spans that will be used for panel
+    :param transparent: defines if the panel is transparent
 
-    :param color: color of threshold
-    :param index: index of color in gauge
-    :param value: when to use this color will be null if index is 0
+    :param showLabels: Show or hide the unique labels column, which shows only non-common labels
+    :param showTime: Show or hide the log timestamp column
+    :param wrapLogMessages: Toggle line wrapping
+    :param sortOrder: Display results in 'Descending' or 'Ascending' time order. The default is Descending,
+        showing the newest logs first.
+    """
+    showLabels = attr.ib(default=False, validator=instance_of(bool))
+    showTime = attr.ib(default=False, validator=instance_of(bool))
+    wrapLogMessages = attr.ib(default=False, validator=instance_of(bool))
+    sortOrder = attr.ib(default='Descending', validator=instance_of(str))
+
+    def to_json_data(self):
+        return self.panel_json(
+            {
+                'fieldConfig': {
+                    'defaults': {
+                        'custom': {},
+                    },
+                    'overrides': []
+                },
+                'options': {
+                    'showLabels': self.showLabels,
+                    'showTime': self.showTime,
+                    'sortOrder': self.sortOrder,
+                    'wrapLogMessage': self.wrapLogMessages
+                },
+                'type': LOGS_TYPE,
+            }
+        )
+
+
+@attr.s
+class Threshold(object):
+    """Threshold for for panels
+    (https://grafana.com/docs/grafana/latest/panels/thresholds/)
+
+    :param color: Color of threshold
+    :param index: Index of color in panel
+    :param line: Display Threshold line, defaults to True
+    :param value: When to use this color will be null if index is 0
+    :param op: EVAL_LT for less than or EVAL_GT for greater than to indicate what the threshold applies to.
+    :param yaxis: Choose left or right for panels
+
+    Example:
+        thresholds = [
+            Threshold('green', 0, 0.0),
+            Threshold('red', 1, 80.0)
+        ]
     """
 
     color = attr.ib()
@@ -2543,6 +2720,54 @@ class Threshold(object):
         }
 
 
+@attr.s
+class GraphThreshold(object):
+    """Threshold for for Graph panel
+    (https://grafana.com/docs/grafana/latest/panels/thresholds/)
+
+    :param colorMode: Color mode of the threshold, value can be `ok`, `warning`, `critical` or `custom`.
+        If `custom` is selcted a lineColor and fillColor should be provided
+    :param fill: Display threshold fill, defaults to True
+    :param line: Display threshold line, defaults to True
+    :param value: When to use this color will be null if index is 0
+    :param op: EVAL_LT for less than or EVAL_GT for greater than to indicate what the threshold applies to.
+    :param yaxis: Choose left or right for Graph panels
+    :param fillColor: Fill color of the threshold, when colorMode = "custom"
+    :param lineColor: Line color of the threshold, when colorMode = "custom"
+
+    Example:
+        thresholds = [
+            GraphThreshold(colorMode="ok", value=10.0),
+            GrpahThreshold(colorMode="critical", value=90.0)
+        ]
+    """
+
+    value = attr.ib(validator=instance_of(float))
+    colorMode = attr.ib(default="critical")
+    fill = attr.ib(default=True, validator=instance_of(bool))
+    line = attr.ib(default=True, validator=instance_of(bool))
+    op = attr.ib(default=EVAL_GT)
+    yaxis = attr.ib(default='left')
+    fillColor = attr.ib(default=RED)
+    lineColor = attr.ib(default=RED)
+
+    def to_json_data(self):
+        data = {
+            'value': self.value,
+            'colorMode': self.colorMode,
+            'fill': self.fill,
+            'line': self.line,
+            'op': self.op,
+            'yaxis': self.yaxis,
+        }
+
+        if self.colorMode == "custom":
+            data['fillColor'] = self.fillColor
+            data['lineColor'] = self.lineColor
+
+        return data
+
+
 @attr.s
 class SeriesOverride(object):
     alias = attr.ib()
